"""The hashing module contains all methods and classes related to the hashing trick."""

import hashlib
import math
import multiprocessing
import platform
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd

import category_encoders.utils as util

__author__ = 'willmcginnis', 'LiuShulun'


class HashingEncoder( util.UnsupervisedTransformerMixin,util.BaseEncoder):
    """A multivariate hashing implementation with configurable dimensionality/precision.

    The advantage of this encoder is that it does not maintain a dictionary of observed categories.
    Consequently, the encoder does not grow in size and accepts new values during data scoring
    by design.

    It's important to read about how max_process & max_sample work
    before setting them manually, inappropriate setting slows down encoding.

    Default value of 'max_process' is 1 on Windows because multiprocessing might cause issues,
    see in :
    https://github.com/scikit-learn-contrib/categorical-encoding/issues/215
    https://docs.python.org/2/library/multiprocessing.html?highlight=process#windows

    Parameters
    ----------
    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform
        (otherwise it will be a numpy array).
    hash_method: str
        which hashing method to use. Any method from hashlib works.
    max_process: int
        how many processes to use in transform(). Limited in range(1, 64).
        By default, it uses half of the logical CPUs.
        For example, 4C4T makes max_process=2, 4C8T makes max_process=4.
        Set it larger if you have a strong CPU.
        It is not recommended to set it larger than is the count of the
        logical CPUs as it will actually slow down the encoding.
    max_sample: int
        how many samples to encode by each process at a time.
        This setting is useful on low memory machines.
        By default, max_sample=(all samples num)/(max_process).
        For example, 4C8T CPU with 100,000 samples makes max_sample=25,000,
        6C12T CPU with 100,000 samples makes max_sample=16,666.
        It is not recommended to set it larger than the default value.
    n_components: int
        how many bits to use to represent the feature. By default, we use 8 bits.
        For high-cardinality features, consider using up-to 32 bits.
    process_creation_method: string
        either "fork", "spawn" or "forkserver" (availability depends on your
        platform). See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        for more details and tradeoffs. Defaults to "fork" on linux/macos as it
        is the fastest option and to "spawn" on windows as it is the only one
        available

    Example
    -------
    >>> from category_encoders.hashing import HashingEncoder
    >>> import pandas as pd
    >>> from sklearn.datasets import fetch_openml
    >>> bunch = fetch_openml(name='house_prices', as_frame=True)
    >>> display_cols = [
    ...     'Id',
    ...     'MSSubClass',
    ...     'MSZoning',
    ...     'LotFrontage',
    ...     'YearBuilt',
    ...     'Heating',
    ...     'CentralAir',
    ... ]
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
    >>> y = bunch.target
    >>> he = HashingEncoder(cols=['CentralAir', 'Heating']).fit(X, y)
    >>> numeric_dataset = he.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 1460 entries, 0 to 1459
    Data columns (total 13 columns):
     #   Column       Non-Null Count  Dtype
    ---  ------       --------------  -----
     0   col_0        1460 non-null   int64
     1   col_1        1460 non-null   int64
     2   col_2        1460 non-null   int64
     3   col_3        1460 non-null   int64
     4   col_4        1460 non-null   int64
     5   col_5        1460 non-null   int64
     6   col_6        1460 non-null   int64
     7   col_7        1460 non-null   int64
     8   Id           1460 non-null   float64
     9   MSSubClass   1460 non-null   float64
     10  MSZoning     1460 non-null   object
     11  LotFrontage  1201 non-null   float64
     12  YearBuilt    1460 non-null   float64
    dtypes: float64(4), int64(8), object(1)
    memory usage: 148.4+ KB
    None

    References
    ----------
    .. [1] Feature Hashing for Large Scale Multitask Learning, from
    https://alex.smola.org/papers/2009/Weinbergeretal09.pdf
    .. [2] Don't be tricked by the Hashing Trick, from
    https://booking.ai/dont-be-tricked-by-the-hashing-trick-192a6aae3087

    """

    prefit_ordinal = False
    encoding_relation = util.EncodingRelation.ONE_TO_M

    def __init__(
        self,
        max_process=0,
        max_sample=0,
        verbose=0,
        n_components=8,
        cols=None,
        drop_invariant=False,
        return_df=True,
        hash_method='md5',
        process_creation_method='fork',
    ):
        super().__init__(
            verbose=verbose,
            cols=cols,
            drop_invariant=drop_invariant,
            return_df=return_df,
            handle_unknown='does not apply',
            handle_missing='does not apply',
        )

        if max_process not in range(1, 128):
            if platform.system() == 'Windows':
                self.max_process = 1
            else:
                self.max_process = int(math.ceil(multiprocessing.cpu_count() / 2))
                if self.max_process < 1:
                    self.max_process = 1
                elif self.max_process > 128:
                    self.max_process = 128
        else:
            self.max_process = max_process
        self.max_sample = int(max_sample)
        if platform.system() == 'Windows':
            self.process_creation_method = 'spawn'
        else:
            self.process_creation_method = process_creation_method
        self.data_lines = 0
        self.X = None

        self.n_components = n_components
        self.hash_method = hash_method

    def _fit(self, X, y=None, **kwargs):
        pass

    def _transform(self, X, override_return_df=False):
        """Perform the transformation to new categorical data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]

        Returns
        -------
        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """
        if self._dim is None:
            raise ValueError('Must train encoder before it can be used to transform data.')

        # first check the type
        X = util.convert_input(X)

        # then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError(f'Unexpected input dimension {X.shape[1]}, expected {self._dim}')

        if not list(self.cols):
            return X

        X = self.hashing_trick(
            X,
            hashing_method=self.hash_method,
            N=self.n_components,
            cols=self.cols,
        )

        return X

    @staticmethod
    def hash_chunk(hash_method: str, np_df: np.ndarray, N: int) -> np.ndarray:
        """Perform hashing on the given numpy array.

        Parameters
        ----------
        hash_method: str
            Hashlib method to use.
        np_df: np.ndarray
            Data to hash.
        N: int
            Number of bits to encode the data.

        Returns
        -------
        np.ndarray
            Hashed data.
        """
        # Calling getattr outside the loop saves some time in the loop
        hasher_constructor = getattr(hashlib, hash_method)
        # Same when the call to getattr is implicit
        int_from_bytes = int.from_bytes
        result = np.zeros((np_df.shape[0], N), dtype='int')
        for i, row in enumerate(np_df):
            for val in row:
                if val is not None:
                    hasher = hasher_constructor()
                    # Computes an integer index from the hasher digest. The endian is
                    # "big" as the code use to read:
                    # column_index = int(hasher.hexdigest(), 16) % N
                    # which is implicitly considering the hexdigest to be big endian,
                    # even if the system is little endian.
                    # Building the index that way is about 30% faster than using the
                    # hexdigest.
                    hasher.update(bytes(str(val), 'utf-8'))
                    column_index = int_from_bytes(hasher.digest(), byteorder='big') % N
                    result[i, column_index] += 1
        return result

    def hashing_trick_with_np_parallel(self, df: pd.DataFrame, N: int) -> pd.DataFrame:
        """Perform the hashing trick in parallel.

        Parameters
        ----------
        df: pd.DataFrame
           data to hash.
        N: int
           how many bits to use to represent the feature.

        Returns
        -------
        pd.DataFrame
           hashed data.
        """
        np_df = df.to_numpy()
        ctx = multiprocessing.get_context(self.process_creation_method)

        with ProcessPoolExecutor(max_workers=self.max_process, mp_context=ctx) as executor:
            result = np.concatenate(
                list(
                    executor.map(
                        self.hash_chunk,
                        [self.hash_method] * self.max_process,
                        np.array_split(np_df, self.max_process),
                        [N] * self.max_process,
                    )
                )
            )

        return pd.DataFrame(result, index=df.index)

    def hashing_trick_with_np_no_parallel(self, df: pd.DataFrame, N: int) -> pd.DataFrame:
        """Perform the hashing trick in a single thread (non-parallel).

        Parameters
        ----------
        df: pd.DataFrame
           data to hash.
        N: int
           how many bits to use to represent the feature.

        Returns
        -------
        pd.DataFrame
           hashed data.
        """
        np_df = df.to_numpy()

        result = HashingEncoder.hash_chunk(self.hash_method, np_df, N)

        return pd.DataFrame(result, index=df.index)

    def hashing_trick(self, X_in, hashing_method='md5', N=2, cols=None, make_copy=False):
        """A basic hashing implementation with configurable dimensionality/precision.

        Performs the hashing trick on a pandas dataframe, `X`, using the hashing method from
        hashlib identified by `hashing_method`.
        The number of output dimensions (`N`), and columns to hash (`cols`) are also configurable.

        Parameters
        ----------
        X_in: pandas dataframe
            description text
        hashing_method: string, optional
            description text
        N: int, optional
            description text
        cols: list, optional
            description text
        make_copy: bool, optional
            description text

        Returns
        -------
        out : dataframe
            A hashing encoded dataframe.

        References
        ----------
        Cite the relevant literature, e.g. [1]_.  You may also cite these
        references in the notes section above.
        .. [1] Kilian Weinberger; Anirban Dasgupta; John Langford; Alex Smola;
        Josh Attenberg (2009). Feature Hashing for Large Scale Multitask Learning. Proc. ICML.

        """
        if hashing_method not in hashlib.algorithms_available:
            raise ValueError(
                f"Hashing Method: {hashing_method} not available. "
                f"Please use one of: {', '.join([str(x) for x in hashlib.algorithms_available])}"
            )

        if make_copy:
            X = X_in.copy(deep=True)
        else:
            X = X_in

        if cols is None:
            cols = X.columns

        new_cols = [f'col_{d}' for d in range(N)]

        X_cat = X.loc[:, cols]
        X_num = X.loc[:, [x for x in X.columns if x not in cols]]

        if self.max_process == 1:
            X_cat = self.hashing_trick_with_np_no_parallel(X_cat, N)
        else:
            X_cat = self.hashing_trick_with_np_parallel(X_cat, N)

        X_cat.columns = new_cols

        X = pd.concat([X_cat, X_num], axis=1)

        return X



"""Target Encoder."""

from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
from scipy.special import expit

import category_encoders.utils as util
from category_encoders.ordinal import OrdinalEncoder

__author__ = 'chappers'


class TargetEncoder( util.SupervisedTransformerMixin,util.BaseEncoder):
    """Target encoding for categorical features.

    Supported targets: binomial and continuous.
    For polynomial target support, see PolynomialWrapper.

    For the case of categorical target: features are replaced with a blend of posterior
    probability of the target given particular categorical value and the prior probability
    of the target over all the training data.

    For the case of continuous target: features are replaced with a blend of the expected value
    of the target given particular categorical value and the expected value of the
    target over all the training data.

    Parameters
    ----------
    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform
        (otherwise it will be a numpy array).
    handle_missing: str
        options are 'error', 'return_nan'  and 'value', defaults to 'value',
        which returns the target mean.
    handle_unknown: str
        options are 'error', 'return_nan' and 'value', defaults to 'value',
        which returns the target mean.
    min_samples_leaf: int
        For regularization the weighted average between category mean and global mean is taken.
        The weight is an S-shaped curve between 0 and 1 with the number of samples for a category
        on the x-axis. The curve reaches 0.5 at min_samples_leaf.
        (parameter k in the original paper)
    smoothing: float
        smoothing effect to balance categorical average vs prior. Higher value means stronger
        regularization. The value must be strictly bigger than 0.
        Higher values mean a flatter S-curve (see min_samples_leaf).
    hierarchy: dict or dataframe
        A dictionary or a dataframe to define the hierarchy for mapping.

        If a dictionary, this contains a dict of columns to map into hierarchies.
        Dictionary key(s) should be the column name from X which requires mapping.
        For multiple hierarchical maps, this should be a dictionary of dictionaries.

        If dataframe: a dataframe defining columns to be used for the hierarchies.
        Column names must take the form:
            HIER_colA_1, ... HIER_colA_N, HIER_colB_1, ... HIER_colB_M, ...
        where [colA, colB, ...] are given columns in cols list.
        1:N and 1:M define the hierarchy for each column where 1 is the highest hierarchy
        (top of the tree).  A single column or multiple can be used, as relevant.

    Examples
    --------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import fetch_openml
    >>> display_cols = [
    ...     'Id',
    ...     'MSSubClass',
    ...     'MSZoning',
    ...     'LotFrontage',
    ...     'YearBuilt',
    ...     'Heating',
    ...     'CentralAir',
    ... ]
    >>> bunch = fetch_openml(name='house_prices', as_frame=True)
    >>> y = bunch.target > 200000
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
    >>> enc = TargetEncoder(cols=['CentralAir', 'Heating'], min_samples_leaf=20, smoothing=10).fit(
    ...     X, y
    ... )
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 1460 entries, 0 to 1459
    Data columns (total 7 columns):
     #   Column       Non-Null Count  Dtype
    ---  ------       --------------  -----
     0   Id           1460 non-null   float64
     1   MSSubClass   1460 non-null   float64
     2   MSZoning     1460 non-null   object
     3   LotFrontage  1201 non-null   float64
     4   YearBuilt    1460 non-null   float64
     5   Heating      1460 non-null   float64
     6   CentralAir   1460 non-null   float64
    dtypes: float64(6), object(1)
    memory usage: 80.0+ KB
    None

    >>> from category_encoders.datasets import load_compass
    >>> X, y = load_compass()
    >>> hierarchical_map = {'compass': {'N': ('N', 'NE'), 'S': ('S', 'SE'), 'W': 'W'}}
    >>> enc = TargetEncoder(
    ...     verbose=1, smoothing=2, min_samples_leaf=2, hierarchy=hierarchical_map, cols=['compass']
    ... ).fit(X.loc[:, ['compass']], y)
    >>> hierarchy_dataset = enc.transform(X.loc[:, ['compass']])
    >>> print(hierarchy_dataset['compass'].values)
    [0.62263617 0.62263617 0.90382995 0.90382995 0.90382995 0.17660024
     0.17660024 0.46051953 0.46051953 0.46051953 0.46051953 0.40332791
     0.40332791 0.40332791 0.40332791 0.40332791]
    >>> X, y = load_postcodes('binary')
    >>> cols = ['postcode']
    >>> HIER_cols = ['HIER_postcode_1', 'HIER_postcode_2', 'HIER_postcode_3', 'HIER_postcode_4']
    >>> enc = TargetEncoder(
    ...     verbose=1, smoothing=2, min_samples_leaf=2, hierarchy=X[HIER_cols], cols=['postcode']
    ... ).fit(X['postcode'], y)
    >>> hierarchy_dataset = enc.transform(X['postcode'])
    >>> print(hierarchy_dataset.loc[0:10, 'postcode'].values)
    [0.75063473 0.90208756 0.88328833 0.77041254 0.68891504 0.85012847
    0.76772574 0.88742357 0.7933824  0.63776756 0.9019973 ]

    References
    ----------

    .. [1] A Preprocessing Scheme for High-Cardinality Categorical Attributes in Classification
    and Prediction Problems, from https://dl.acm.org/citation.cfm?id=507538

    """

    prefit_ordinal = True
    encoding_relation = util.EncodingRelation.ONE_TO_ONE

    def __init__(
        self,
        verbose: int = 0,
        cols: list[str] = None,
        drop_invariant: bool = False,
        return_df: bool = True,
        handle_missing: str = 'value',
        handle_unknown: str = 'value',
        min_samples_leaf: int = 20,
        smoothing: float = 10,
        hierarchy: dict = None,
    ) -> None:
        super().__init__(
            verbose=verbose,
            cols=cols,
            drop_invariant=drop_invariant,
            return_df=return_df,
            handle_unknown=handle_unknown,
            handle_missing=handle_missing,
        )
        self.ordinal_encoder = None
        self.min_samples_leaf = min_samples_leaf
        self.smoothing = smoothing
        self.mapping = None
        self._mean = None
        # @ToDo create a function to check the hierarchy
        if isinstance(hierarchy, (dict, pd.DataFrame)) and cols is None:
            raise ValueError('Hierarchy is defined but no columns are named for encoding')
        if isinstance(hierarchy, dict):
            self.hierarchy = {}
            self.hierarchy_depth = {}
            for switch in hierarchy:
                flattened_hierarchy = util.flatten_reverse_dict(hierarchy[switch])
                hierarchy_check = self._check_dict_key_tuples(flattened_hierarchy)
                self.hierarchy_depth[switch] = hierarchy_check[1]
                if not hierarchy_check[0]:
                    raise ValueError(
                        'Hierarchy mapping contains different levels for key "' + switch + '"'
                    )
                self.hierarchy[switch] = {
                    (k if isinstance(t, tuple) else t): v
                    for t, v in flattened_hierarchy.items()
                    for k in t
                }
        elif isinstance(hierarchy, pd.DataFrame):
            self.hierarchy = hierarchy
            self.hierarchy_depth = {}
            for col in self.cols:
                HIER_cols = self.hierarchy.columns[
                    self.hierarchy.columns.str.startswith(f'HIER_{col}')
                ].tolist()
                HIER_levels = [int(i.replace(f'HIER_{col}_', '')) for i in HIER_cols]
                if np.array_equal(sorted(HIER_levels), np.arange(1, max(HIER_levels) + 1)):
                    self.hierarchy_depth[col] = max(HIER_levels)
                else:
                    raise ValueError(f'Hierarchy columns are not complete for column {col}')
        elif hierarchy is None:
            self.hierarchy = hierarchy
        else:
            raise ValueError('Given hierarchy mapping is neither a dictionary nor a dataframe')

        self.cols_hier = []

    @staticmethod
    def _check_dict_key_tuples(dict_to_check: dict[Any, tuple]) -> tuple[bool, int]:
        """Check if all tuples in the dict values have the same length.

        Parameters
        ----------
        dict_to_check: dictionary to check

        Returns
        -------
        tuple: first entry if all sizes are equal, second minimum size.
        """
        min_tuple_size = min(len(v) for v in dict_to_check.values())
        max_tuple_size = max(len(v) for v in dict_to_check.values())
        return min_tuple_size == max_tuple_size, min_tuple_size

    def _fit(self, X: util.X_type, y: util.y_type, **kwargs) -> None:
        if isinstance(self.hierarchy, dict):
            X_hier = pd.DataFrame()
            for switch in self.hierarchy:
                if switch in self.cols:
                    colnames = [
                        f'HIER_{str(switch)}_{str(i + 1)}'
                        for i in range(self.hierarchy_depth[switch])
                    ]
                    df = pd.DataFrame(
                        X[str(switch)].map(self.hierarchy[str(switch)]).tolist(),
                        index=X.index,
                        columns=colnames,
                    )
                    X_hier = pd.concat([X_hier, df], axis=1)
        elif isinstance(self.hierarchy, pd.DataFrame):
            X_hier = self.hierarchy

        if isinstance(self.hierarchy, (dict, pd.DataFrame)):
            enc_hier = OrdinalEncoder(
                verbose=self.verbose,
                cols=X_hier.columns,
                handle_unknown='value',
                handle_missing='value',
            )
            enc_hier = enc_hier.fit(X_hier)
            X_hier_ordinal = enc_hier.transform(X_hier)

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose, cols=self.cols, handle_unknown='value', handle_missing='value'
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)
        if self.hierarchy is not None:
            self.mapping = self.fit_target_encoding(
                pd.concat([X_ordinal, X_hier_ordinal], axis=1), y
            )
        else:
            self.mapping = self.fit_target_encoding(X_ordinal, y)

    def fit_target_encoding(self, X: util.X_type, y: util.y_type) -> dict[str, np.ndarray]:
        """Fit the target encoding mapping.

        Parameters
        ----------
        X: training data to fit on.
        y: training target.

        Returns
        -------
        dictionary: column -> encoding values for column

        """
        mapping = {}
        prior = self._mean = y.mean()

        for switch in self.ordinal_encoder.category_mapping:
            col = switch.get('col')
            if 'HIER_' not in str(col):
                values = switch.get('mapping')

                scalar = prior
                if (isinstance(self.hierarchy, dict) and col in self.hierarchy) or (
                    isinstance(self.hierarchy, pd.DataFrame)
                ):
                    for i in range(self.hierarchy_depth[col]):
                        col_hier = 'HIER_' + str(col) + '_' + str(i + 1)
                        col_hier_m1 = (
                            col
                            if i == self.hierarchy_depth[col] - 1
                            else 'HIER_' + str(col) + '_' + str(i + 2)
                        )
                        if not X[col].equals(X[col_hier]) and len(X[col_hier].unique()) > 1:
                            stats_hier = y.groupby(X[col_hier]).agg(['count', 'mean'])
                            smoove_hier = self._weighting(stats_hier['count'])
                            scalar_hier = (
                                scalar * (1 - smoove_hier) + stats_hier['mean'] * smoove_hier
                            )
                            scalar_hier_long = X[[col_hier_m1, col_hier]].drop_duplicates()
                            scalar_hier_long.index = np.arange(1, scalar_hier_long.shape[0] + 1)
                            scalar = scalar_hier_long[col_hier].map(scalar_hier.to_dict())

                stats = y.groupby(X[col]).agg(['count', 'mean'])
                smoove = self._weighting(stats['count'])

                smoothing = scalar * (1 - smoove) + stats['mean'] * smoove

                if self.handle_unknown == 'return_nan':
                    smoothing.loc[-1] = np.nan
                elif self.handle_unknown == 'value':
                    smoothing.loc[-1] = prior

                if self.handle_missing == 'return_nan':
                    smoothing.loc[values.loc[np.nan]] = np.nan
                elif self.handle_missing == 'value':
                    smoothing.loc[-2] = prior

                mapping[col] = smoothing

        return mapping

    def _transform(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        # Now X is the correct dimensions it works with pre fitted ordinal encoder
        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == 'error':
            if X[self.cols].isin([-1]).any().any():
                raise ValueError('Unexpected categories found in dataframe')

        X = self.target_encode(X)
        return X

    def target_encode(self, X_in: pd.DataFrame) -> pd.DataFrame:
        """Apply target encoding via encoder mapping."""
        X = X_in.copy(deep=True)

        # Was not mapping extra columns as self.featuer_names_in did not include new column
        for col in self.cols:
            X[col] = X[col].map(self.mapping[col])

        return X

    def _weighting(self, n: int) -> float:
        # monotonically increasing function of n bounded between 0 and 1
        # sigmoid in this case, using scipy.expit for numerical stability
        return expit((n - self.min_samples_leaf) / self.smoothing)



"""Count Encoder."""

from copy import copy

import numpy as np
import pandas as pd

import category_encoders.utils as util
from category_encoders.ordinal import OrdinalEncoder

__author__ = 'joshua t. dunn'


class CountEncoder( util.UnsupervisedTransformerMixin,util.BaseEncoder):
    """Count encoding for categorical features.

    For a given categorical feature, replace the names of the groups with the group counts.
    """

    prefit_ordinal = True
    encoding_relation = util.EncodingRelation.ONE_TO_ONE

    def __init__(
        self,
        verbose=0,
        cols=None,
        drop_invariant=False,
        return_df=True,
        handle_unknown='value',
        handle_missing='value',
        min_group_size=None,
        combine_min_nan_groups=None,
        min_group_name=None,
        normalize=False,
    ):
        """Count encoding for categorical features.

        For a given categorical feature, replace the names of the groups
        with the group counts.

        Parameters
        ----------
        verbose: int
            integer indicating verbosity of output. 0 for none.
        cols: list
            a list of columns to encode, if None, all string and categorical columns
            will be encoded.
        drop_invariant: bool
            boolean for whether or not to drop columns with 0 variance.
        return_df: bool
            boolean for whether to return a pandas DataFrame from transform
            (otherwise it will be a numpy array).
        handle_missing: str
            how to handle missing values at fit time. Options are 'error', 'return_nan',
            and 'value'. Default 'value', which treat nans as a countable category at
            fit time.
        handle_unknown: str, int or dict of {column : option, ...}.
            how to handle unknown labels at transform time. Options are 'error'
            'return_nan', 'value' and int. Defaults to None which uses nan behaviour
            specified at fit time. Passing an int will fill with this int value.
        normalize: bool or dict of {column : bool, ...}.
            whether to normalize the counts to the range (0, 1). See Pandas `value_counts`
            for more details.
        min_group_size: int, float or dict of {column : option, ...}.
            the minimal count threshold of a group needed to ensure it is not
            combined into a "leftovers" group. Default value is 0.01.
            If float in the range (0, 1),
            `min_group_size` is calculated as int(X.shape[0] * min_group_size).
            Note: This value may change type based on the `normalize` variable. If True
            this will become a float. If False, it will be an int.
        min_group_name: None, str or dict of {column : option, ...}.
            Set the name of the combined minimum groups when the defaults become
            too long. Default None. In this case the category names will be joined
            alphabetically with a `_` delimiter.
            Note: The default name can be long and may keep changing, for example,
            in cross-validation.
        combine_min_nan_groups: bool or dict of {column : bool, ...}.
            whether to combine the leftovers group with nan group. Default True. Can
            also be forced to combine with 'force' meaning small groups are effectively
            counted as nans. Force can only be used when 'handle_missing' is 'value' or 'error'.
            Note: Will not force if it creates a binary or invariant column.


        Example
        -------
        >>> import pandas as pd
        >>> from sklearn.datasets import fetch_openml
        >>> from category_encoders import CountEncoder

        >>> bunch = fetch_openml(name='house_prices', as_frame=True)
        >>> display_cols = [
        ...     'Id',
        ...     'MSSubClass',
        ...     'MSZoning',
        ...     'LotFrontage',
        ...     'YearBuilt',
        ...     'Heating',
        ...     'CentralAir',
        ... ]
        >>> y = bunch.target
        >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
        >>> enc = CountEncoder(cols=['CentralAir', 'Heating']).fit(X, y)
        >>> numeric_dataset = enc.transform(X)
        >>> print(numeric_dataset.info())
        <class 'pandas.core.frame.DataFrame'>
        RangeIndex: 1460 entries, 0 to 1459
        Data columns (total 7 columns):
         #   Column       Non-Null Count  Dtype
        ---  ------       --------------  -----
         0   Id           1460 non-null   float64
         1   MSSubClass   1460 non-null   float64
         2   MSZoning     1460 non-null   object
         3   LotFrontage  1201 non-null   float64
         4   YearBuilt    1460 non-null   float64
         5   Heating      1460 non-null   int64
         6   CentralAir   1460 non-null   int64
        dtypes: float64(4), int64(2), object(1)
        memory usage: 80.0+ KB
        None
        """
        super().__init__(
            verbose=verbose,
            cols=cols,
            drop_invariant=drop_invariant,
            return_df=return_df,
            handle_unknown=handle_unknown,
            handle_missing=handle_missing,
        )
        self.mapping = None
        self.normalize = normalize
        self.min_group_size = min_group_size
        self.min_group_name = min_group_name
        self.combine_min_nan_groups = combine_min_nan_groups
        self.ordinal_encoder = None

        self._check_set_create_attrs()

        self._min_group_categories = {}
        self._normalize = {}
        self._min_group_name = {}
        self._combine_min_nan_groups = {}
        self._min_group_size = {}
        self._handle_unknown = {}
        self._handle_missing = {}

    def _fit(self, X, y=None, **kwargs):
        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose, cols=self.cols, handle_unknown='value', handle_missing='value'
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)

        self._check_set_create_dict_attrs()
        self._fit_count_encode(X_ordinal, y)
        return self

    def _transform(self, X):
        for col in self.cols:
            # Treat None as np.nan
            X[col] = pd.Series(
                [el if el is not None else np.nan for el in X[col]], index=X[col].index
            )
            if self.handle_missing == 'value':
                if not util.is_category(X[col].dtype):
                    X[col] = X[col].fillna(np.nan)

            if self._min_group_size is not None:
                if col in self._min_group_categories.keys():
                    X[col] = X[col].map(self._min_group_categories[col]).fillna(X[col])

            X[col] = X[col].astype(object).map(self.mapping[col])
            if isinstance(self._handle_unknown[col], (int, np.integer)):
                X[col] = X[col].fillna(self._handle_unknown[col])

            elif (
                self._handle_unknown[col] == 'value'
                and X[col].isna().any()
                and self._handle_missing[col] != 'return_nan'
            ):
                X[col] = X[col].replace(np.nan, 0)

            elif self._handle_unknown[col] == 'error' and X[col].isna().any():
                raise ValueError(f'Missing data found in column {col} at transform time.')
        return X

    def _fit_count_encode(self, X_in, y):
        """Perform the count encoding."""
        X = X_in.copy(deep=True)

        if self.cols is None:
            self.cols = X.columns

        self.mapping = {}

        for col in self.cols:
            mapping_values = X[col].value_counts(normalize=self._normalize[col])
            ordinal_encoding = [
                m['mapping'] for m in self.ordinal_encoder.mapping if m['col'] == col
            ][0]
            reversed_ordinal_enc = {v: k for k, v in ordinal_encoding.to_dict().items()}
            mapping_values.index = mapping_values.index.map(reversed_ordinal_enc)
            self.mapping[col] = mapping_values

            if self._handle_missing[col] == 'return_nan':
                self.mapping[col][np.nan] = np.nan

            # elif self._handle_missing[col] == 'value':
            # test_count.py failing     self.mapping[col].loc[-2] = 0

        if any(val is not None for val in self._min_group_size.values()):
            self.combine_min_categories(X)

    def combine_min_categories(self, X):
        """Combine small categories into a single category."""
        for col, mapper in self.mapping.items():
            if self._normalize[col] and isinstance(self._min_group_size[col], int):
                self._min_group_size[col] = self._min_group_size[col] / X.shape[0]
            elif not self._normalize and isinstance(self._min_group_size[col], float):
                self._min_group_size[col] = self._min_group_size[col] * X.shape[0]

            if self._combine_min_nan_groups[col] is True:
                min_groups_idx = mapper < self._min_group_size[col]
            elif self._combine_min_nan_groups[col] == 'force':
                min_groups_idx = (mapper < self._min_group_size[col]) | (mapper.index.isna())
            else:
                min_groups_idx = (mapper < self._min_group_size[col]) & (~mapper.index.isna())

            min_groups_sum = mapper.loc[min_groups_idx].sum()

            if (
                min_groups_sum > 0
                and min_groups_idx.sum() > 1
                and not min_groups_idx.loc[~min_groups_idx.index.isna()].all()
            ):
                if isinstance(self._min_group_name[col], str):
                    min_group_mapper_name = self._min_group_name[col]
                else:
                    min_group_mapper_name = '_'.join(
                        [
                            str(idx)
                            for idx in mapper.loc[min_groups_idx].index.astype(str).sort_values()
                        ]
                    )

                self._min_group_categories[col] = {
                    cat: min_group_mapper_name for cat in mapper.loc[min_groups_idx].index.tolist()
                }

                if not min_groups_idx.all():
                    mapper = mapper.loc[~min_groups_idx]
                    mapper[min_group_mapper_name] = min_groups_sum

                self.mapping[col] = mapper

    def _check_set_create_attrs(self):
        """Check attributes setting that don't play nicely `self.cols`."""
        if not (
            (self.combine_min_nan_groups in ['force', True, False, None])
            or isinstance(self.combine_min_nan_groups, dict)
        ):
            raise ValueError(
                "'combine_min_nan_groups' should be one of: "
                "['force', True, False, None] or type dict."
            )

        if self.handle_missing == 'return_nan' and self.combine_min_nan_groups == 'force':
            raise ValueError(
                "Cannot have `handle_missing` == 'return_nan' and "
                "'combine_min_nan_groups' == 'force' for all columns."
            )

        if self.combine_min_nan_groups is not None and self.min_group_size is None:
            pass
            # raise ValueError(
            #     "`combine_min_nan_groups` only works when `min_group_size` "
            #     "is set for all columns."
            # )

        if self.min_group_name is not None and self.min_group_size is None:
            raise ValueError(
                '`min_group_name` only works when `min_group_size` is set ' 'for all columns.'
            )

        if self.combine_min_nan_groups is None:
            self.combine_min_nan_groups = True

    def _check_set_create_dict_attrs(self):
        """Check attributes that can be dicts and format for all `self.cols`."""
        dict_attrs = {
            'normalize': False,
            'min_group_name': None,
            'combine_min_nan_groups': True,
            'min_group_size': None,
            'handle_unknown': 'value',
            'handle_missing': 'value',
        }

        for attr_name, attr_default in dict_attrs.items():
            attr = copy(getattr(self, attr_name))
            if isinstance(attr, dict):
                for col in self.cols:
                    if col not in attr:
                        attr[col] = attr_default
                setattr(self, '_' + attr_name, attr)
            else:
                attr_dict = {}
                for col in self.cols:
                    attr_dict[col] = attr
                setattr(self, '_' + attr_name, attr_dict)

        for col in self.cols:
            if (
                self._handle_missing[col] == 'return_nan'
                and self._combine_min_nan_groups[col] == 'force'
            ):
                raise ValueError(
                    "Cannot have `handle_missing` == 'return_nan' and "
                    f"'combine_min_nan_groups' == 'force' for columns `{col}`."
                )

            if self._combine_min_nan_groups[col] is not True and self._min_group_size[col] is None:
                raise ValueError(
                    f'`combine_min_nan_groups` only works when `min_group_size` is set for '
                    f'column {col}.'
                )

            if self._min_group_name[col] is not None and self._min_group_size[col] is None:
                raise ValueError(
                    f'`min_group_name` only works when `min_group_size` is set for '
                    f'column {col}.'
                )



"""Module for wrappers that add extra functionality to encoders."""

from __future__ import annotations

import copy

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import StratifiedKFold

import category_encoders as encoders
from category_encoders import utils


class PolynomialWrapper(BaseEstimator, TransformerMixin):
    """Extend supervised encoders to n-class labels, where n >= 2.

    The label can be numerical (e.g.: 0, 1, 2, 3,...,n), string or categorical (pandas.Categorical).
    The label is first encoded into n-1 binary columns. Subsequently, the inner supervised encoder
    is executed for each binarized label.

    The names of the encoded features are suffixed with underscore and the corresponding class name
    (edge scenarios like 'dog'+'cat_frog' vs. 'dog_cat'+'frog' are not currently handled).

    The implementation is experimental and the API may change in the future.
    The order of the returned features may change in the future.


    Parameters
    ----------
    feature_encoder: Object
        an instance of a supervised encoder.


    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import fetch_openml
    >>> from category_encoders.wrapper import PolynomialWrapper
    >>> display_cols = [
    ...     'Id',
    ...     'MSSubClass',
    ...     'MSZoning',
    ...     'LotFrontage',
    ...     'YearBuilt',
    ...     'Heating',
    ...     'CentralAir',
    ... ]
    >>> bunch = fetch_openml(name='house_prices', as_frame=True)
    >>> # need more than one column
    >>> y = bunch.target.map(lambda x: int(min([x, 300000]) / 50000))
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
    >>> enc = TargetEncoder(cols=['CentralAir', 'Heating'])
    >>> wrapper = PolynomialWrapper(enc)
    >>> encoded = wrapper.fit_transform(X, y)
    >>> print(encoded.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 1460 entries, 0 to 1459
    Data columns (total 17 columns):
     #   Column        Non-Null Count  Dtype
    ---  ------        --------------  -----
     0   Id            1460 non-null   float64
     1   MSSubClass    1460 non-null   float64
     2   MSZoning      1460 non-null   object
     3   LotFrontage   1201 non-null   float64
     4   YearBuilt     1460 non-null   float64
     5   CentralAir_3  1460 non-null   float64
     6   Heating_3     1460 non-null   float64
     7   CentralAir_2  1460 non-null   float64
     8   Heating_2     1460 non-null   float64
     9   CentralAir_5  1460 non-null   float64
     10  Heating_5     1460 non-null   float64
     11  CentralAir_6  1460 non-null   float64
     12  Heating_6     1460 non-null   float64
     13  CentralAir_1  1460 non-null   float64
     14  Heating_1     1460 non-null   float64
     15  CentralAir_0  1460 non-null   float64
     16  Heating_0     1460 non-null   float64
    dtypes: float64(16), object(1)
    memory usage: 194.0+ KB
    None
    """

    def __init__(self, feature_encoder: utils.BaseEncoder):
        """Init the polynomial wrapper."""
        self.feature_encoder: utils.BaseEncoder = feature_encoder
        self.feature_encoders: dict[str, utils.BaseEncoder] = {}
        self.label_encoder: encoders.OneHotEncoder | None = None

    def fit(self, X, y, **kwargs):
        """Fit a multi-label encoder."""
        # unite the input into pandas types
        X, y = utils.convert_inputs(X, y)
        y = pd.DataFrame(y.rename('target'))

        # apply one-hot-encoder on the label
        self.label_encoder = encoders.OneHotEncoder(
            handle_missing='error',
            handle_unknown='error',
            cols=['target'],
            drop_invariant=True,
            use_cat_names=True,
        )
        labels = self.label_encoder.fit_transform(y)
        labels.columns = [column[7:] for column in labels.columns]
        labels = labels.iloc[:, 1:]  # drop one label

        # train the feature encoders, it is important to reset feature encoders first
        self.feature_encoders = {}
        for class_name, label in labels.items():
            self.feature_encoders[class_name] = copy.deepcopy(self.feature_encoder).fit(X, label)

    def transform(self, X, y=None):
        """Encode new data."""
        # unite the input into pandas types
        X = utils.convert_input(X)

        # initialization
        encoded = None
        feature_encoder = None
        all_new_features = pd.DataFrame()

        # transform the features
        if y is not None:
            y = self.label_encoder.transform(pd.DataFrame({'target': y}))
        for class_name, feature_encoder in self.feature_encoders.items():
            if y is not None:
                y_transform = y[f'target_{class_name}']
            else:
                y_transform = None
            encoded = feature_encoder.transform(X, y_transform)

            # decorate the encoded features with the label class suffix
            new_features = encoded[feature_encoder.cols]
            new_features.columns = [
                str(column) + '_' + class_name for column in new_features.columns
            ]

            all_new_features = pd.concat((all_new_features, new_features), axis=1)

        # add features that were not encoded
        result = pd.concat(
            (
                encoded[encoded.columns[~encoded.columns.isin(feature_encoder.cols)]],
                all_new_features,
            ),
            axis=1,
        )

        return result

    def fit_transform(self, X, y=None, **fit_params):
        """Fit encoder and encode the training data."""
        self.fit(X, y, **fit_params)
        return self.transform(X, y)


class NestedCVWrapper(BaseEstimator, TransformerMixin):
    """Cross validate supervised encoders to avoid overfitting.

    For a validation or a test set, supervised encoders can be used as follows:

        X_train_encoded = encoder.fit_transform(X_train, y_train)
        X_valid_encoded = encoder.transform(X_valid)

    However, the downstream model will be overfitting to the encoded training data due to
    target leakage. Using out-of-fold encodings is an effective way to prevent target leakage.
    This is equivalent to:

        X_train_encoded = np.zeros(X.shape)
        for trn, val in kfold.split(X, y):
            encoder.fit(X[trn], y[trn])
            X_train_encoded[val] = encoder.transform(X[val])

    This can be used in place of the "inner folds" as discussed here:
        https://sebastianraschka.com/faq/docs/evaluate-a-model.html

    See README.md for a list of supervised encoders.

    Discussion: Although leave-one-out encoder internally performs leave-one-out cross-validation,
    it is actually the most overfitting supervised model in our library.
    To illustrate the issue, let's imagine we have a totally unpredictive nominal feature
    and a perfectly balanced binary label. A supervised encoder should encode the feature into a
    constant vector as the feature is unpredictive of the label.
    But when we use leave-one-out cross-validation, the label ratio cease to be perfectly balanced
    and the wrong class label always becomes the majority in the training fold.
    Leave-one-out encoder returns a seemingly predictive feature.
    And the downstream model starts to overfit to the encoded feature.
    Unfortunately, even 10-fold cross-validation is not immune to this effect:
        http://www.kdd.org/exploration_files/v12-02-4-UR-Perlich.pdf
    To decrease the effect, it is recommended to use a low count of the folds.
    And that is the reason why this wrapper uses 5 folds by default.

    Based on the empirical results, only LeaveOneOutEncoder benefits greatly from this wrapper.
    The remaining encoders can be used without this wrapper.


    Parameters
    ----------
    feature_encoder: Object
        an instance of a supervised encoder.

    cv: int or sklearn cv Object
        if an int is given, StratifiedKFold is used by default, where the int
        is the number of folds.

    shuffle: boolean, optional
        whether to shuffle each classes samples before splitting into batches.
        Ignored if a CV method is provided.

    random_state: int, RandomState instance or None, optional, default=None
        if int, random_state is the seed used by the random number generator.
        Ignored if a CV method is provided.


    Example
    -------
    >>> import pandas as pd
    >>> from category_encoders import *
    >>> from category_encoders.wrapper import NestedCVWrapper
    >>> from sklearn.datasets import fetch_openml
    >>> from sklearn.model_selection import GroupKFold, train_test_split
    >>> bunch = fetch_openml(name='house_prices', as_frame=True)
    >>> display_cols = [
    ...     'Id',
    ...     'MSSubClass',
    ...     'MSZoning',
    ...     'LotFrontage',
    ...     'YearBuilt',
    ...     'Heating',
    ...     'CentralAir',
    ... ]
    >>> y = bunch.target > 200000
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
    >>> X_train, X_test, y_train, _ = train_test_split(X, y, random_state=42)
    >>> X_train, X_valid, y_train, y_valid = train_test_split(X_train, y_train, random_state=42)
    >>> # Define the nested CV encoder for a supervised encoder
    >>> enc_nested = NestedCVWrapper(TargetEncoder(cols=['CentralAir', 'Heating']), random_state=42)
    >>> # Encode the X data for train, valid & test
    >>> X_train_enc, X_valid_enc, X_test_enc = enc_nested.fit_transform(
    ...     X_train, y_train, X_test=(X_valid, X_test)
    ... )
    >>> print(X_train_enc.info())
    <class 'pandas.core.frame.DataFrame'>
    Int64Index: 821 entries, 1390 to 896
    Data columns (total 7 columns):
     #   Column       Non-Null Count  Dtype
    ---  ------       --------------  -----
     0   Id           821 non-null    float64
     1   MSSubClass   821 non-null    float64
     2   MSZoning     821 non-null    object
     3   LotFrontage  672 non-null    float64
     4   YearBuilt    821 non-null    float64
     5   Heating      821 non-null    float64
     6   CentralAir   821 non-null    float64
    dtypes: float64(6), object(1)
    memory usage: 51.3+ KB
    None
    """

    def __init__(self, feature_encoder, cv=5, shuffle=True, random_state=None):
        """Init wrapper."""
        self.feature_encoder = feature_encoder
        self.__name__ = feature_encoder.__class__.__name__
        self.shuffle = shuffle
        self.random_state = random_state

        if isinstance(cv, int):
            self.cv = StratifiedKFold(n_splits=cv, shuffle=shuffle, random_state=random_state)
        else:
            self.cv = cv

    def fit(self, X, y, **kwargs):
        """Fit on the base feature_encoder without nested cross validation."""
        self.feature_encoder.fit(X, y, **kwargs)

    def transform(self, X):
        """Transform on the base feature_encoder without nested cross validation."""
        return self.feature_encoder.transform(X)

    def fit_transform(self, X, y=None, X_test=None, groups=None, **fit_params):
        """Unbiased encodings from a supervised encoder and inference on test set.

        :param X: array-like, shape = [n_samples, n_features]
                  Training vectors for the supervised encoder,
                  where n_samples is the number of samples
                  and n_features is the number of features.
        :param y: array-like, shape = [n_samples]
                  Target values for the supervised encoder.
        :param X_test, optional: array-like, shape = [m_samples, n_features]
                       or a tuple of array-likes (X_test, X_valid...)
                       Vectors to be used for inference by an encoder
                       (e.g. test or validation sets) trained on the
                       full X & y sets. No nested folds are used here
        :param groups: Groups to be passed to the cv method, e.g. for GroupKFold
        :param fit_params:
        :return: array, shape = [n_samples, n_numeric + N]
                 Transformed values with encoding applied.
                 Returns multiple arrays if X_test is not None
        """
        X, y = utils.convert_inputs(X, y)

        # Get out-of-fold encoding for the training data
        out_of_fold = pd.DataFrame()

        for trn_idx, oof_idx in self.cv.split(X, y, groups):
            feature_encoder = copy.deepcopy(self.feature_encoder)
            feature_encoder.fit(X.iloc[trn_idx], y.iloc[trn_idx])
            out_of_fold = pd.concat([out_of_fold, feature_encoder.transform(X.iloc[oof_idx])])

        # Train the encoder on all the training data for testing data
        self.feature_encoder = copy.deepcopy(self.feature_encoder)
        self.feature_encoder.fit(X, y)

        if X_test is None:
            return out_of_fold
        else:
            if isinstance(X_test, tuple):
                encoded_data = (out_of_fold,)
                for dataset in X_test:
                    encoded_data = encoded_data + (self.feature_encoder.transform(dataset),)
                return encoded_data
            else:
                return out_of_fold, self.feature_encoder.transform(X_test)



"""Backward difference contrast encoding."""

import numpy as np
from patsy.contrasts import ContrastMatrix, Diff

from category_encoders.base_contrast_encoder import BaseContrastEncoder

__author__ = 'paulwestenthanner'


class BackwardDifferenceEncoder(BaseContrastEncoder):
    """Backward difference contrast coding for encoding categorical variables.

    Parameters
    ----------
    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform
        (otherwise it will be a numpy array).
    handle_unknown: str
        options are 'error', 'return_nan', 'value', and 'indicator'. The default is 'value'.
        Warning: if indicator is used, an extra column will be added in if the transform matrix
        has unknown categories. This can cause unexpected changes in dimension in some cases.
    handle_missing: str
        options are 'error', 'return_nan', 'value', and 'indicator'. The default is 'value'.
        Warning: if indicator is used, an extra column will be added in if the transform
        matrix has nan values. This can cause unexpected changes in dimension in some cases.

    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import fetch_openml
    >>> bunch = fetch_openml(name='house_prices', as_frame=True)
    >>> display_cols = [
    ...     'Id',
    ...     'MSSubClass',
    ...     'MSZoning',
    ...     'LotFrontage',
    ...     'YearBuilt',
    ...     'Heating',
    ...     'CentralAir',
    ... ]
    >>> y = bunch.target
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
    >>> enc = BackwardDifferenceEncoder(cols=['CentralAir', 'Heating']).fit(X, y)
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 1460 entries, 0 to 1459
    Data columns (total 12 columns):
     #   Column        Non-Null Count  Dtype
    ---  ------        --------------  -----
     1   Id            1460 non-null   float64
     2   MSSubClass    1460 non-null   float64
     3   MSZoning      1460 non-null   object
     4   LotFrontage   1201 non-null   float64
     5   YearBuilt     1460 non-null   float64
     6   Heating_0     1460 non-null   float64
     7   Heating_1     1460 non-null   float64
     8   Heating_2     1460 non-null   float64
     9   Heating_3     1460 non-null   float64
     10  Heating_4     1460 non-null   float64
     11  CentralAir_0  1460 non-null   float64
    dtypes: float64(10), int64(1), object(1)
    memory usage: 137.0+ KB
    None

    References
    ----------

    .. [1] Contrast Coding Systems for Categorical Variables, from
    https://stats.idre.ucla.edu/r/library/r-library-contrast-coding-systems-for-categorical-variables/

    .. [2] Gregory Carey (2003). Coding Categorical Variables, from
    http://ibgwww.colorado.edu/~carey/p5741ndir/Coding_Categorical_Variables.pdf

    """

    def get_contrast_matrix(self, values_to_encode: np.array) -> ContrastMatrix:
        """Get the contrast matrix for the backward difference encoder."""
        return Diff().code_without_intercept(values_to_encode)



"""A collection of shared utilities for all encoders, not intended for external use."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from dataclasses import dataclass, fields
from enum import Enum, auto
from typing import Hashable, Sequence

import numpy as np
import pandas as pd
import sklearn.base
from pandas.api.types import is_numeric_dtype, is_object_dtype, is_string_dtype
from pandas.core.dtypes.dtypes import CategoricalDtype
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import Tags

__author__ = 'willmcginnis'

X_type = np.ndarray | pd.DataFrame | list | np.generic | csr_matrix
y_type = list | pd.Series | np.ndarray | tuple | pd.DataFrame


def convert_cols_to_list(
    cols: pd.Series | np.ndarray | set | tuple | CategoricalDtype | str | int,
) -> list:
    """Convert columns to list.

    Parameters
    ----------
    cols: columns as Series, array, set, tuple, ...

    Returns
    -------
    columns as list.

    """
    if isinstance(cols, pd.Series):
        return cols.tolist()
    elif isinstance(cols, np.ndarray):
        return cols.tolist()
    elif np.isscalar(cols):
        return [cols]
    elif isinstance(cols, set):
        return list(cols)
    elif isinstance(cols, tuple):
        return list(cols)
    elif isinstance(cols, CategoricalDtype):
        return cols.astype(object).tolist()

    return cols


def get_categorical_cols(df: pd.DataFrame) -> list[str]:
    """Returns names of categorical columns in the DataFrame.

    These include columns of types: object, category, string, string[pyarrow].

    Parameters
    ----------
    df DataFrame

    Returns
    -------
    list of columns

    """
    obj_cols = []
    for col, dtype in df.dtypes.items():
        if is_object_dtype(dtype) or is_category(dtype) or is_string_dtype(dtype):
            # if not isinstance(col, str):
            #     raise ValueError(f'DataFrame column names must be strings not {col}.')
            obj_cols.append(col)

    if not obj_cols:
        msg = (
            'Warning: No categorical columns found. '
            "Calling 'transform' will only return input data."
        )
        print(msg)

    return obj_cols


def is_category(dtype: pd.core.dtypes.dtypes.ExtensionDtype) -> bool:
    """Check if dtype is pandas categorical type.

    Parameters
    ----------
    dtype pandas dtype

    Returns
    -------
    True if CategoricalDtype, False otherwise.

    """
    return isinstance(dtype, CategoricalDtype)


def convert_inputs(
    X: X_type,
    y: y_type | None,
    columns: Sequence = None,
    index: Sequence = None,
    deep: bool = False,
) -> tuple[pd.DataFrame, pd.Series | None]:
    """
    Unite arraylike `X` and vectorlike `y` into a DataFrame and Series.

    If both are pandas types already, raises an error if their indexes do not match.
    If one is pandas, the returns will share that index.
    If neither is pandas, a default index will be used, unless `index` is passed.

    Parameters
    ----------
    X: arraylike
    y: listlike
    columns: listlike
        Specifies column names to use for `X`.
        Ignored if `X` is already a dataframe.
        If `None`, use the default pandas column names.
    index: listlike
        The index to use, if neither `X` nor `y` is a pandas type.
        (If one has an index, then this has no effect.)
        If `None`, use the default pandas index.
    deep: bool
        Whether to deep-copy `X`.
    """
    X_alt_index = y.index if isinstance(y, pd.Series) else index
    X = convert_input(X, columns=columns, deep=deep, index=X_alt_index)
    if y is not None:
        y = convert_input_vector(y, index=X.index)
        # N.B.: If either was already pandas, it keeps its index.

        if any(X.index != y.index):
            msg = (
                '`X` and `y` both have indexes, but they do not match. If you are shuffling '
                'your input data on purpose (e.g. via permutation_test_score) use '
                'np arrays instead of data frames / series'
            )
            raise ValueError(msg)
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                'The length of X is '
                + str(X.shape[0])
                + ' but length of y is '
                + str(y.shape[0])
                + '.'
            )
    return X, y


def convert_input(
    X: X_type, columns: Sequence = None, index: Sequence = None, deep: bool = False
) -> pd.DataFrame:
    """Unite data into a DataFrame.

    Objects that do not contain column names take the names from the argument.
    Optionally perform deep copy of the data.

    Parameters
    ----------
    X: data
    columns: column names to assign, ignored if data is already a data frame.
    index: index to use for the dataframe. Defaults to range(len(data)).
    deep: flag whether the data should be copied when creating the data frame.

    Returns
    -------
    A dataframe with the data and columns and index properly set.
    """
    if not isinstance(X, pd.DataFrame):
        if isinstance(X, pd.Series):
            X = pd.DataFrame(X, copy=deep)
        else:
            if columns is not None and np.size(X, 1) != len(columns):
                raise ValueError(
                    'The count of the column names does not correspond to the count of the columns'
                )
            if isinstance(X, list):
                X = pd.DataFrame(
                    X, columns=columns, copy=deep, index=index
                )  # lists are always copied, but for consistency, we still pass the argument
            elif isinstance(X, (np.generic, np.ndarray)):
                X = pd.DataFrame(X, columns=columns, copy=deep, index=index)
            elif isinstance(X, csr_matrix):
                X = pd.DataFrame(X.todense(), columns=columns, copy=deep, index=index)
            else:
                raise ValueError(f'Unexpected input type: {type(X)}')
    elif deep:
        X = X.copy(deep=True)

    return X


def convert_input_vector(y: y_type, index: Sequence) -> pd.Series:
    """Unite target data type into a Series.

    If the target is a Series or a DataFrame, we preserve its index.
    But if the target does not contain index attribute, we use the index from the argument.

    Parameters
    ----------
    y: target data to convert to series.
    index: index to be used for the series.

    Returns
    -------
    pd.Series containing the target.

    """
    if y is None:
        raise ValueError(
            'Supervised encoders need a target for the fitting. The target cannot be None'
        )
    if isinstance(y, pd.Series):
        return y
    elif isinstance(y, np.ndarray):
        if len(np.shape(y)) == 1:  # vector
            return pd.Series(y, name='target', index=index)
        elif len(np.shape(y)) == 2 and np.shape(y)[0] == 1:  # single row in a matrix
            return pd.Series(y[0, :], name='target', index=index)
        elif len(np.shape(y)) == 2 and np.shape(y)[1] == 1:  # single column in a matrix
            return pd.Series(y[:, 0], name='target', index=index)
        else:
            raise ValueError(f'Unexpected input shape: {np.shape(y)}')
    elif np.isscalar(y):
        raise ValueError('y must be a list, an np.ndarray or a pd.Series. Not a scalar')
    elif isinstance(y, list):
        if len(y) == 0:  # empty list
            return pd.Series(y, name='target', index=index, dtype=float)
        elif len(y) > 0 and not isinstance(y[0], list):  # vector
            return pd.Series(y, name='target', index=index)
        elif len(y) > 0 and isinstance(y[0], list) and len(y[0]) == 1:  # single row in a matrix

            def flatten(y):
                return [item for sublist in y for item in sublist]

            return pd.Series(flatten(y), name='target', index=index)
        elif (
            len(y) == 1 and len(y[0]) == 0 and isinstance(y[0], list)
        ):  # single empty column in a matrix
            return pd.Series(y[0], name='target', index=index, dtype=float)
        elif len(y) == 1 and isinstance(y[0], list):  # single column in a matrix
            return pd.Series(y[0], name='target', index=index, dtype=type(y[0][0]))
        else:
            raise ValueError('Unexpected input shape')
    elif isinstance(y, pd.DataFrame):
        if len(list(y)) == 0:  # empty DataFrame
            return pd.Series(name='target', index=index, dtype=float)
        if len(list(y)) == 1:  # a single column
            return y.iloc[:, 0]
        else:
            raise ValueError(f'Unexpected input shape: {y.shape}')
    else:
        return pd.Series(
            y, name='target', index=index
        )  # this covers tuples and other directly convertible types


def get_generated_cols(
    X_original: pd.DataFrame, X_transformed: pd.DataFrame, to_transform: list[Hashable]
) -> list[Hashable]:
    """
    Returns a list of the generated/transformed columns.

    Arguments:
        X_original: df
            the original (input) DataFrame.
        X_transformed: df
            the transformed (current) DataFrame.
        to_transform: [str]
            a list of columns that were transformed (as in the original DataFrame),
            commonly self.feature_names_in.

    Output:
        a list of columns that were transformed (as in the current DataFrame).
    """
    original_cols = list(X_original.columns)

    if len(to_transform) > 0:
        [original_cols.remove(c) for c in to_transform]

    current_cols = list(X_transformed.columns)
    if len(original_cols) > 0:
        [current_cols.remove(c) for c in original_cols]

    return current_cols


def flatten_reverse_dict(dict_to_flatten: dict) -> dict:
    """Flatten a dictionary into a tuple of nested keys.

    Parameters
    ----------
    dict_to_flatten

    Returns
    -------
    the flattened dictionary with tuples as keys indicating the hierarchy.

    """
    sep = '___'
    [flat_dict] = pd.json_normalize(dict_to_flatten, sep=sep).to_dict(orient='records')
    reversed_flat_dict = {v: tuple(k.split(sep)) for k, v in flat_dict.items()}
    return reversed_flat_dict


class EncodingRelation(Enum):
    """Relation of how many input features are encoded into how many output features."""

    # one input feature get encoded into one output feature
    ONE_TO_ONE = auto()
    # one input feature get encoded into as many output features as it has distinct values
    ONE_TO_N_UNIQUE = auto()
    # one input feature get encoded into m output features
    # that are not the number of distinct values
    ONE_TO_M = auto()
    # all N input features are encoded into M output features.
    # The encoding is done globally on all the input not on a per-feature basis
    N_TO_M = auto()


def get_docstring_output_shape(in_out_relation: EncodingRelation) -> str:
    """Find how many encoded features are expected.

    Parameters
    ----------
    in_out_relation

    Returns
    -------
    A string saying how many features to expect.

    """
    if in_out_relation == EncodingRelation.ONE_TO_ONE:
        return 'n_features'
    elif in_out_relation == EncodingRelation.ONE_TO_N_UNIQUE:
        return 'n_features * respective cardinality'
    elif in_out_relation == EncodingRelation.ONE_TO_M:
        return 'M features (n_features < M)'
    elif in_out_relation == EncodingRelation.N_TO_M:
        return 'M features (M can be anything)'


@dataclass
class EncoderTags(Tags):
    """Custom Tags for encoders."""

    predict_depends_on_y: bool = False

    @classmethod
    def from_sk_tags(cls, tags: Tags) -> EncoderTags:
        """Initialize EncoderTags from given sklearn Tags."""
        as_dict = {
            field.name: getattr(tags, field.name)
            for field in fields(tags)
        }
        return cls(**as_dict)

class BaseEncoder(BaseEstimator):
    """BaseEstimator class for all encoders.

    This follows the sklearn estimator / transformer pattern.
    """

    _dim: int | None
    cols: list[str]
    use_default_cols: bool
    handle_missing: str
    handle_unknown: str
    verbose: int
    drop_invariant: bool
    invariant_cols: list[str] = []
    return_df: bool
    supervised: bool
    encoding_relation: EncodingRelation

    INVARIANCE_THRESHOLD = (
        10e-5  # columns with variance less than this will be considered constant / invariant
    )

    def __init__(
        self,
        verbose: int = 0,
        cols: list[str] = None,
        drop_invariant: bool = False,
        return_df: bool = True,
        handle_unknown: str = 'value',
        handle_missing: str = 'value',
        **kwargs,
    ):
        """Initialize the encoder.

        Parameters
        ----------
        verbose: int
            integer indicating verbosity of output. 0 for none.
        cols: list
            a list of columns to encode, if None, all string and categorical columns
            will be encoded.
        drop_invariant: bool
            boolean for whether to drop columns with 0 variance.
        return_df: bool
            boolean for whether to return a pandas DataFrame from transform and inverse transform
            (otherwise it will be a numpy array).
        handle_missing: str
            how to handle missing values at fit time. Options are 'error', 'return_nan',
            and 'value'. Default 'value', which treat nans as a countable category at
            fit time.
        handle_unknown: str, int or dict of {column : option, ...}.
            how to handle unknown labels at transform time. Options are 'error'
            'return_nan', 'value' and int. Defaults to None which uses nan behaviour
            specified at fit time. Passing an int will fill with this int value.
        kwargs: dict.
            additional encoder specific parameters like regularisation.
        """
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.invariant_cols = []
        self.verbose = verbose
        # if True, even a repeated call of fit() will select string columns from X
        self.use_default_cols = cols is None
        # note that cols are only the columns to be encoded, feature_names_in_ are all columns
        self.cols = cols
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._dim = None

    def fit(self, X: X_type, y: y_type | None = None, **kwargs):
        """Fits the encoder according to X and y.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target values.

        Returns
        -------
        self : encoder
            Returns self.

        """
        X, y = convert_inputs(X, y)
        self._check_fit_inputs(X, y)
        self.feature_names_in_ = X.columns.tolist()
        self.n_features_in_ = len(self.feature_names_in_)

        if self.__sklearn_tags__().target_tags.required:
            if not is_numeric_dtype(y):
                self.lab_encoder_ = LabelEncoder()
                y = self.lab_encoder_.fit_transform(y)
            else:
                self.lab_encoder_ = None

        self._dim = X.shape[1]
        self._determine_fit_columns(X)

        if not set(self.cols).issubset(X.columns):
            raise ValueError('X does not contain the columns listed in cols')

        if self.handle_missing == 'error':
            if X[self.cols].isna().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        self._fit(X, y, **kwargs)

        # for finding invariant columns transform without y (as is done on the test set)
        self.feature_names_out_ = None  # Issue#437
        X_transformed = self.transform(X, override_return_df=True)
        self.feature_names_out_ = X_transformed.columns.to_numpy()

        # drop all output columns with 0 variance.
        if self.drop_invariant:
            generated_cols = get_generated_cols(X, X_transformed, self.cols)
            self.invariant_cols = [
                x for x in generated_cols if X_transformed[x].var() <= self.INVARIANCE_THRESHOLD
            ]
            self.feature_names_out_ = np.fromiter(
                (x for x in self.feature_names_out_ if x not in self.invariant_cols),
                dtype=self.feature_names_out_.dtype,
            )

        return self

    def _check_fit_inputs(self, X: X_type, y: y_type) -> None:
        if self.__sklearn_tags__().target_tags.required:
            if y is None:
                raise ValueError(
                    'Supervised encoders need a target for the fitting. The target cannot be None'
                )
            else:
                if y.isna().any():  # Target column should never have missing values
                    raise ValueError('The target column y must not contain missing values.')

    def _check_transform_inputs(self, df: pd.DataFrame) -> None:
        if self.handle_missing == 'error':
            if df[self.cols].isna().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        if self._dim is None:
            raise NotFittedError('Must train encoder before it can be used to transform data.')

        # then make sure that it is the right size
        if df.shape[1] != self._dim:
            raise ValueError(f'Unexpected input dimension {df.shape[1]}, expected {self._dim}')

    def _drop_invariants(
        self, df: pd.DataFrame, override_return_df: bool
    ) -> np.ndarray | pd.DataFrame:
        if self.drop_invariant:
            df = df.drop(columns=self.invariant_cols)

        if self.return_df or override_return_df:
            return df
        else:
            return df.to_numpy()

    def _determine_fit_columns(self, X: pd.DataFrame) -> None:
        """Determine columns used by encoder.

        Note that the implementation also deals with re-fitting the same encoder object
        with different columns.

        :param X: input data frame
        :return: none, sets self.cols as a side effect
        """
        # if columns aren't passed, just use every string column
        if self.use_default_cols:
            self.cols = get_categorical_cols(X)
        else:
            self.cols = convert_cols_to_list(self.cols)

    def get_feature_names(self) -> np.ndarray:
        """Deprecated method to get feature names. Use `get_feature_names_out` instead."""
        msg = (
            '`get_feature_names` is deprecated in all of sklearn. '
            'Use `get_feature_names_out` instead.'
        )
        warnings.warn(msg, category=FutureWarning, stacklevel=2)
        return self.get_feature_names_out()

    def get_feature_names_out(self, input_features=None) -> np.ndarray:
        """Get the names of all transformed / added columns.

        Note that in sklearn the get_feature_names_out function takes the feature_names_in
        as an argument and determines the output feature names using the input.
        A fit is usually not necessary and if so a NotFittedError is raised.
        We just require a fit all the time and return the fitted output columns.

        Returns
        -------
        feature_names: np.ndarray
            A numpy array with all feature names transformed or added.
            Note: potentially dropped features (because the feature is constant/invariant)
            are not included!

        """
        out_feats = getattr(self, 'feature_names_out_', None)
        if not isinstance(out_feats, np.ndarray):
            raise NotFittedError('Estimator has to be fitted to return feature names.')
        else:
            return out_feats

    def get_feature_names_in(self) -> np.ndarray:
        """Get the names of all input columns present when fitting.

        These columns are necessary for the transform step.
        """
        in_feats = getattr(self, 'feature_names_in_', None)
        if isinstance(in_feats, list):
            in_feats = np.array(in_feats)
        if not isinstance(in_feats, np.ndarray):
            raise NotFittedError('Estimator has to be fitted to return feature names.')
        else:
            return in_feats

    @abstractmethod
    def _fit(self, X: pd.DataFrame, y: pd.Series | None, **kwargs): ...


class SupervisedTransformerMixin(sklearn.base.TransformerMixin):
    """Mixin for supervised transformers (with target)."""

    def __sklearn_tags__(self) -> EncoderTags:
        """Set scikit transformer tags."""
        sk_tags = super().__sklearn_tags__()
        tags = EncoderTags.from_sk_tags(sk_tags)
        tags.target_tags.required = True
        return tags

    def transform(self, X: X_type, y: y_type | None = None, override_return_df: bool = False):
        """Perform the transformation to new categorical data.

        Some encoders behave differently on whether y is given or not.
        This is mainly due to regularisation in order to avoid overfitting.
        On training data transform should be called with y, on test data without.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        y : array-like, shape = [n_samples] or None
        override_return_df : bool
            override self.return_df to force to return a data frame

        Returns
        -------
        p : array or DataFrame, shape = [n_samples, n_features_out]
            Transformed values with encoding applied.

        """
        # first check the type
        X, y = convert_inputs(X, y, deep=True)
        self._check_transform_inputs(X)
        if y is not None and self.lab_encoder_ is not None:
            y = self.lab_encoder_.transform(y)

        if not list(self.cols):
            return X

        X = self._transform(X, y)

        return self._drop_invariants(X, override_return_df)

    @abstractmethod
    def _transform(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame: ...

    def fit_transform(self, X: X_type, y: y_type | None = None, **fit_params):
        """Fit and transform using the target information.

        This also uses the target for transforming, not only for training.
        """
        if y is None:
            raise TypeError('fit_transform() missing argument: ' 'y' '')
        return self.fit(X, y, **fit_params).transform(X, y)


class UnsupervisedTransformerMixin(sklearn.base.TransformerMixin):
    """Mixin for Transformers without target information."""

    def transform(self, X: X_type, override_return_df: bool = False):
        """Perform the transformation to new categorical data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        override_return_df : bool
            override self.return_df to force to return a data frame

        Returns
        -------
        p : array or DataFrame, shape = [n_samples, n_features_out]
            Transformed values with encoding applied.

        """
        # first check the type
        X = convert_input(X, deep=True)
        self._check_transform_inputs(X)

        if not list(self.cols):
            return X

        X = self._transform(X)
        return self._drop_invariants(X, override_return_df)

    @abstractmethod
    def _transform(self, X: pd.DataFrame) -> pd.DataFrame: ...
